package nl.nictiz.digsig;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.security.Key;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.PrivateKey;
import java.security.Provider;
import java.security.Security;
import java.security.UnrecoverableKeyException;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.util.Enumeration;
import java.util.Properties;


import sun.security.pkcs11.SunPKCS11;

/**
 * This class hides the handling of the PKCS#11 Keystore for the examples. It
 * also offers some convenience methods for the PINcode and loading the
 * configuration.
 */
public class SmartCard {

	/**
	 * Some constants
	 */
	public static final String PKCS11_PROVIDER_TYPE = "PKCS11";
	static final String PKCS_PROVIDER_INSTANCE_NAME = "PKCS";

	/**
	 * Which keystore and certificate will be used. Initialized after the
	 * Keystore is loaded.
	 */
	private KeyStore keystore = null;
	private String authenticationAlias = null;
	private String nonRepudiationAlias = null;

	/**
	 * Create an accessor for the smartcard, given the smartcard's user's pin
	 * code.
	 * @param pincode the user pincode of the smartcard.
	 *   Uses a char array, since Strings might leak to easily to other parts of
	 *   the memory. This has to deal with the JVM internals and memory /
	 *   performance optimizations for Strings.
	 * @param properties These properties contain the "configuration" to access
	 *   the smartcard. This should contain the library name of the PKCS#11
	 *   driver and the number of the "slot" to use. @link
	 *   http://java.sun.com/j2se/1.5.0/docs/guide/security/p11guide.html#Config
	 * @throws IOException 
	 * @throws FileNotFoundException 
	 * @throws CertificateException 
	 * @throws NoSuchAlgorithmException 
	 * @throws KeyStoreException 
	 */
	public SmartCard(char[] userPin, Properties properties)
		throws KeyStoreException, NoSuchAlgorithmException,
			CertificateException, FileNotFoundException, IOException
	{
		this.keystore = getKeyStore(userPin, properties);
	}

	/**
	 * Open the keystore that contains the key and signature to use. Loads the
	 * proper 'alias' to use in this instance.
	 * @return
	 * @throws KeyStoreException
	 * @throws IOException
	 * @throws NoSuchAlgorithmException
	 * @throws CertificateException
	 * @throws FileNotFoundException
	 */
	private KeyStore getKeyStore(char[] pincode, Properties pkcsConfiguration)
		throws KeyStoreException, IOException, NoSuchAlgorithmException,
			CertificateException, FileNotFoundException
	{

		// The smart card is only available if the PKCS crypto (JCE) provider
		// is loaded.
		loadSecurityProvider(pkcsConfiguration);

		// Create the keystore
		KeyStore result = KeyStore.getInstance(PKCS11_PROVIDER_TYPE);
		result.load(null, pincode);

		// Find the proper key alias (certificate for signing)
		// In java, each certificate or key(pair) is accessed under an 'alias' 
		Enumeration<String> aliases = result.aliases();
		while (aliases.hasMoreElements()) {
			String alias = aliases.nextElement();
//			System.err.println(alias);
			if (result.getCertificate(alias) instanceof X509Certificate) {
				X509Certificate certificate =
					(X509Certificate) result.getCertificate(alias);
				// Check that he proper certificate is used
				// The first bit is set on certificates for digital signature
				// See http://java.sun.com/j2se/1.5.0/docs/api/java/security/cert/X509Certificate.html#getKeyUsage()
				if (certificate.getKeyUsage()[0]) {
					this.authenticationAlias = alias;
					System.err.println("Found authentication (TA) certificate = " + alias);
					System.err.println("Issuer: " + certificate.getIssuerX500Principal());
					System.err.println("Dec serial#: " + certificate.getSerialNumber().toString(10));
				}
				if (certificate.getKeyUsage()[1]) {
					this.nonRepudiationAlias = alias;
					System.err.println("Found non-repudiation (EH) certificate = " + alias);
					System.err.println("Issuer: " + certificate.getIssuerX500Principal());
					System.err.println("Dec serial#: " + certificate.getSerialNumber().toString(10));
				}
			}
//			else {
//				System.err.println("No applicable certificate found");
//			}
		}

//		KeyStore keyStore = KeyStore.getInstance("JKS");
//		keyStore.load(new FileInputStream("test.jks"), "blabla".toCharArray());

		return result;
	}

	/**
	 * Load the security provider that offers the proper JCE services for using
	 * the smart card.
	 * @throws IOException
	 */
	private void loadSecurityProvider(Properties pkcsConfiguration)
		throws IOException
	{

		Properties prop = new Properties();
		prop.putAll(pkcsConfiguration);
		// Set / override the name, since it is required and we might need it
		prop.setProperty("name", PKCS_PROVIDER_INSTANCE_NAME);

		InputStream providerProperties = propertiesToStream(prop);
		Provider pkcsProvider = new SunPKCS11(providerProperties);

		// Make this provider the first preferred one, so it is also used for
		// digest and crypto algorithms, as well as the key material.
		Security.insertProviderAt(pkcsProvider, 1);

	}

	/**
	 * Make a set of properties available as an (in memory) stream.
	 * @param properties the Properties to convert
	 * @return a InputStream containing the given properties 
	 * @throws IOException
	 */
	private InputStream propertiesToStream(Properties properties)
		throws IOException
	{
		ByteArrayOutputStream buffer = new ByteArrayOutputStream();
		properties.store(buffer, null);
		InputStream result =
			new ByteArrayInputStream(buffer.toByteArray());
		return result;
	}

	/**
	 * Load the effective private key for authentication (TA) from the keystore
	 * @param keyStore
	 * @return the PrivateKey; an object that refers to the key on the card
	 * @throws KeyStoreException
	 * @throws NoSuchAlgorithmException
	 * @throws UnrecoverableKeyException
	 */
	public PrivateKey getAuthenticationKey()
		throws KeyStoreException, NoSuchAlgorithmException,
			UnrecoverableKeyException
	{
		// There is no pincode needed, since the keystore itself was already
		// loaded with the pincode.
		Key key = keystore.getKey(authenticationAlias, null);
		return (PrivateKey) key;
	}

	/**
	 * Load the effective authentication (TA) certificate from the keystore.
	 * @param keyStore
	 * @return
	 * @throws KeyStoreException
	 */
	public X509Certificate getAuthenticationCertificate()
		throws KeyStoreException
	{
		X509Certificate result = 
			(X509Certificate) keystore.getCertificate(authenticationAlias);
		return result;
	}

	/**
	 * Get the authentication (TA) certificate's alias, as the Java Keystore
	 * will know it.
	 * Needed for WSS4J
	 * @return the alias of the certificate to use
	 */
	public String getAuthenticationCertificateAlias() {
		return authenticationAlias;
	}

	/**
	 * Load the effective private key for non-repudiation (EH) from the keystore
	 * @param keyStore
	 * @return the PrivateKey; an object that refers to the key on the card
	 * @throws KeyStoreException
	 * @throws NoSuchAlgorithmException
	 * @throws UnrecoverableKeyException
	 */
	public PrivateKey getNonRepudiationKey()
		throws KeyStoreException, NoSuchAlgorithmException,
			UnrecoverableKeyException
	{
		// There is no pincode needed, since the keystore itself was already
		// loaded with the pincode.
		Key key = keystore.getKey(nonRepudiationAlias, null);
		return (PrivateKey) key;
	}

	/**
	 * Load the effective certificate for non-repudiation (EH) from the keystore
	 * @param keyStore
	 * @return
	 * @throws KeyStoreException
	 */
	public X509Certificate getNonRepudiationCertificate()
		throws KeyStoreException
	{
		X509Certificate result = 
			(X509Certificate) keystore.getCertificate(nonRepudiationAlias);
		return result;
	}

	/**
	 * Get the non-repudiation (EH) certificate's alias, as the Java Keystore
	 * will know it.
	 * Needed for WSS4J
	 * @return the alias of the certificate to use
	 */
	public String getNonRepudiationCertificateAlias() {
		return nonRepudiationAlias;
	}

	/**
	 * Load the PKCS#11 property file. This contains the configuration to
	 * access the smartcard: which driver, which "slot", etc...
	 * @return a Properties to be used with the SmartCard constructor
	 * @throws FileNotFoundException
	 * @throws IOException
	 */
	public static Properties loadPkcsConfiguration()
		throws FileNotFoundException, IOException
	{
		Properties properties = new Properties();
		FileInputStream configFile = new FileInputStream("pkcs.properties");
		properties.load(configFile);
		return properties;
	}

	/**
	 * Ask the user for the smart card's pincode.
	 * <EM>This method is NOT SAFE, it displays the pincode on screen. Do not
	 * use for production purposes!</EM>
	 * <B>NOTE: do not use with production cards!</B>
	 * @return a pincode
	 * @throws IOException
	 */
	public static char[] readPassword()
		throws IOException
	{
		Reader reader = new InputStreamReader(System.in);
		BufferedReader lineReader = new BufferedReader(reader);
		System.err.print("Please enter your test card's pincode: ");
		String pincode = lineReader.readLine();
		reader.close();
		return pincode.toCharArray();
	}

}
