package nl.nictiz.digsig.ta;

import java.io.IOException;
import java.math.BigInteger;
import java.security.InvalidAlgorithmParameterException;
import java.security.Key;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.PrivateKey;
import java.security.UnrecoverableKeyException;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Properties;

import javax.security.auth.x500.X500Principal;
import javax.xml.crypto.MarshalException;
import javax.xml.crypto.dsig.SignedInfo;
import javax.xml.crypto.dsig.XMLSignature;
import javax.xml.crypto.dsig.XMLSignatureException;
import javax.xml.crypto.dsig.XMLSignatureFactory;
import javax.xml.crypto.dsig.dom.DOMSignContext;
import javax.xml.crypto.dsig.keyinfo.KeyInfo;
import javax.xml.crypto.dsig.keyinfo.KeyInfoFactory;
import javax.xml.crypto.dsig.keyinfo.X509Data;
import javax.xml.crypto.dsig.keyinfo.X509IssuerSerial;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerException;

import nl.nictiz.digsig.SmartCard;
import nl.nictiz.digsig.XmlCommon;
import nl.nictiz.digsig.XmlLibrarySignature;

import org.w3c.dom.Document;
import org.xml.sax.SAXException;

/**
 * An example how to create WS-Security signatures using a XML-Signature
 * library.
 */
public class TAXmlLibrarySignature extends XmlLibrarySignature {

	/**
	 * @param args
	 * @throws CertificateException 
	 * @throws NoSuchAlgorithmException 
	 * @throws KeyStoreException 
	 * @throws UnrecoverableKeyException 
	 * @throws SAXException 
	 * @throws ParserConfigurationException 
	 * @throws TransformerException 
	 * @throws XMLSignatureException 
	 * @throws MarshalException 
	 * @throws InvalidAlgorithmParameterException 
	 */
	public static void main(String[] args)
		throws IOException, KeyStoreException, NoSuchAlgorithmException,
			CertificateException, UnrecoverableKeyException,
			ParserConfigurationException, SAXException,
			InvalidAlgorithmParameterException, MarshalException,
			XMLSignatureException, TransformerException
	{
		// switch SHA-256 mode based on program startup argument
		boolean sha256mode = false;
		List<String> arguments = new ArrayList<String>(Arrays.asList(args));
		if (arguments.contains("-sha256")) {
			arguments.remove("-sha256");
			sha256mode = true;
		}

		// Which document to sign?
		String filename = arguments.get(0);
		Document document = XmlCommon.getDocument(filename);

		// Get the configuration to use
		Properties properties = SmartCard.loadPkcsConfiguration();

		// Access the smartcard
		char[] pincode = SmartCard.readPassword();
		SmartCard smartCard = new SmartCard(pincode, properties);

		// key will be a reference to the key on the smartcard, not the actual
		// key itself. Other parts of the libraries can cope with this "key".
		PrivateKey key = smartCard.getAuthenticationKey();
		X509Certificate certificate = smartCard.getAuthenticationCertificate();

		// Add the signature to the document
		createSignature(document, key, certificate,
				"#token_2.16.528.1.1007.3.3.1234567.1_0123456789", sha256mode);

		// NOTE KeyInfo in WS-Security differs from plain XML-Signature: it has
		// an extra element wss:SecurityTokenReference between KeyInfo and
		// X509IssuerSerial. That is:
		//    <KeyInfo><X509Data>....
		// versus
		//    <KeyInfo><wss:SecurityTokenReference><X509Data>...
		// Both the WSS4J and Metro implementations of WS-Security use
		// internals of their related XML-Security library to accomplish this.
		// A pure java.xml.crypto solution does not seem possible, without
		// implementing a custom KeyInfoFactory :-(
		Document wssDocument =
			XmlCommon.transform(document, "ta-xmlsignature-to-wss.xsl");

		// display the result
		XmlCommon.displayXml(wssDocument);

	}


	/**
	 * Create a XML-signature
	 * @param document the document to contain the signature
	 * @param key the key for the signature
	 * @param certificate the X509certificate that identifies the owner of
	 *     the key (i.e. smart-card)
	 * @param nodeId the ID of the node to sign in the document
	 * @throws NoSuchAlgorithmException
	 * @throws InvalidAlgorithmParameterException
	 * @throws MarshalException
	 * @throws XMLSignatureException
	 * @throws TransformerException 
	 * @throws IOException 
	 */
	protected static void createSignature(Document document, Key key,
			X509Certificate certificate, String nodeId, boolean sha256mode)
	    throws NoSuchAlgorithmException, InvalidAlgorithmParameterException,
	    	MarshalException, XMLSignatureException, TransformerException,
	    	IOException
	{
		// Some Java boilerplate
		XMLSignatureFactory signatureFactory =
			XMLSignatureFactory.getInstance("DOM");

		// The information about the signature itself
		SignedInfo signedInfo =
			makeSignedInfo(signatureFactory, nodeId, sha256mode);

		// The information about the signer
		KeyInfo keyInfo = makeKeyInfo(certificate);

		// Create the actual signature
		DOMSignContext signContext =
			new DOMSignContext(key, document.getDocumentElement());
		XMLSignature signature =
			signatureFactory.newXMLSignature(signedInfo, keyInfo);
		signature.sign(signContext);

//		showCanonicalizedForm(signedInfo);

	}

	/**
	 * Create the information about the key used to create the signature
	 * @param certificate
	 * @return
	 */
	protected static KeyInfo makeKeyInfo(X509Certificate certificate) {

		KeyInfoFactory keyInfoFactory = KeyInfoFactory.getInstance();

		// Send the issuer and the certificate's serialnumber
		X500Principal issuerX500Principal = certificate.getIssuerX500Principal();
		BigInteger serialNumber = certificate.getSerialNumber();
		X509IssuerSerial issuerSerial =
			keyInfoFactory.newX509IssuerSerial(issuerX500Principal.toString(), serialNumber);

		X509Data keyInfoData =
			keyInfoFactory.newX509Data(Collections.singletonList(issuerSerial));

		KeyInfo keyInfo =
			keyInfoFactory.newKeyInfo(Collections.singletonList(keyInfoData));

		return keyInfo;
	}

}
